#include "std_includes.h"
#include "network.h"
#include "sysfs.h"
#include "endianness.h"

extern int interface_count;
extern struct interface_context interfaces[50];

int report_to_kernel = 1;
int monitor_writing_to_sysfs = 0;

int verbose = 1;

//internal flags:
int increment = 1;
int show_hex = 0;
int monitor_link_state =1;
int promisc_mode = 0;
int multicast_mode = 1;
int fast_sleep = 0;
int assert = 1;
int show_errors = 0;

char *path_to_sysfs = 0;
	
void print_hex(unsigned char *buf, int length)
{
	int i;
	for (i = 0; i < length; i++) {
		if ((i%16 == 0) && (i != 0)) printf("\n");
		printf("%.2x", buf[i]&0xff);
		if (i%2) printf(" ");
	}

	printf("\n");
}

int init_dcbx_specific_fields(struct interface_context *ifc)
{
	ifc->dirty = 1;
	ifc->link_state = -1;
	ifc->announced_create = 0;
	ifc->seqno = 1;
	ifc->ackno = 0;
	ifc->fcoe_cfg = 0;
	ifc->priority_flow_control = 0;
	ifc->tell_kernel_link_changed = 0;
	ifc->saw_an_fcoe_application_tlv = 0;
	return 0;
}

void change_link_state(struct interface_context *ifc, int new_state)
{
	if (ifc->link_state != new_state) {
		ifc->link_state = new_state;
		if (verbose)
			printf("%s: link is %s.\n", ifc->name, ifc->link_state?"up":"down");
		ifc->tell_kernel_link_changed = 1;
	}
}

/*******************************************************/
/*                 Send Packet                         */
/*******************************************************/
int write_tl(unsigned char* buf, unsigned char type, int length)
{
	buf[0] = (type<<1) + ((length>>8)&0x1);
	buf[1] = length&0xff;
	return 2;
}

int write_tlv(unsigned char* buffer, unsigned char type, int length, unsigned char* value)
{
	unsigned char *buf;
	int i;

	buf = buffer;
	buf += write_tl(buf, type, length);
	
	for (i = 0; i < length; i++) {
		*(buf++) = value[i];
	}

	return buf - buffer;
}

int write_control_tlv(unsigned char *buf, struct protocol_control *pc)
{
	buf[0] = pc->oper_version;
	buf[1] = pc->max_version;
	buf[2] = ((pc->seqno)>>24)&0xff;
	buf[3] = ((pc->seqno)>>16)&0xff;
	buf[4] = ((pc->seqno)>>8)&0xff;
	buf[5] = ((pc->seqno)>>0)&0xff;

	buf[6] = ((pc->ackno)>>24)&0xff;
	buf[7] = ((pc->ackno)>>16)&0xff;
	buf[8] = ((pc->ackno)>>8)&0xff;
	buf[9] = ((pc->ackno)>>0)&0xff;

	return 10;
}

int write_feature_tlv(unsigned char *buf, struct feature *fe)
{
	int i;

	buf[0] = fe->oper_version;
	buf[1] = fe ->max_version;
	buf[2] = 0;
	if (fe->enable) buf[2] |= 0x80;
	if (fe->willing) buf[2] |= 0x40;
	if (fe->error) buf[2] |= 0x20;
	buf[3] = fe->sub_type&0xff;

	for (i = 0; i < fe->cfg_size; i++) {
		buf[4 + i] = fe->cfg[i];
	}
	return 4 + i;
}

int write_osp(unsigned char *buffer, struct packet *p)
{
	unsigned char *buf;
	int i, s;

	buf = buffer;
	*(buf++) = 0x00;
	*(buf++) = 0x1b;
	*(buf++) = 0x21;
	*(buf++) = 0x01;

	s = write_control_tlv(buf + 2, &(p->pc));
	buf += write_tl(buf, 1, s);
	buf += s;

	for (i = 0; i < p->fe_size; i++) {
		s = write_feature_tlv(buf + 2,&(p->fe[i]));
		buf += write_tl(buf, p->fe[i].type, s);
		buf += s;
	}

	return buf - buffer;
}

int send_packet(struct interface_context *ifc)
{
	unsigned char buffer[1000];
	unsigned char *buf;
	unsigned char temp[1000];
	int i, s;

	buf = buffer;
	buf += write_tlv(buf, 1, 7, (ifc->pkt.lh.chassis_id));
	buf += write_tlv(buf, 2, 7, (ifc->pkt.lh.port_id));

	temp[0] = ((ifc->pkt.lh.ttl)>>8) & 0xff;
	temp[1] = (ifc->pkt.lh.ttl) & 0xff;
	buf += write_tlv(buf, 3, 2, temp);

	s = write_osp(buf + 2, &(ifc->pkt));
	buf += write_tl(buf, 127, s);
	buf += s;

	buf += write_tlv(buf, 0, 0, 0);

	send_buffer(ifc, buffer, buf - buffer);
	return 0;
}

void init_packet(struct interface_context *ifc)
{
	int i, fn;
	ifc->pkt.lh.chassis_id[0] = 0x04;

	for(i = 0; i < 6; i++)
		ifc->pkt.lh.chassis_id[1 + i]= ifc->mac[i];

	ifc->pkt.lh.port_id[0] = 0x03;

	for(i = 0; i < 6; i++)
		ifc->pkt.lh.port_id[1 + i]= ifc->mac[i];

	ifc->pkt.lh.ttl = 0x78;
	
	ifc->pkt.pc.oper_version = 0;
	ifc->pkt.pc.max_version = 0;
	ifc->pkt.pc.seqno = ifc->seqno;
	ifc->pkt.pc.ackno = ifc->ackno;

	fn = 0;

	ifc->pkt.fe[fn].type = 5;
	ifc->pkt.fe[fn].oper_version = 0;
	ifc->pkt.fe[fn].max_version = 0;
	ifc->pkt.fe[fn].enable = 1;
	ifc->pkt.fe[fn].willing = 1;
	ifc->pkt.fe[fn].error = 0;
	ifc->pkt.fe[fn].sub_type = 0;
	ifc->pkt.fe[fn].cfg_size = 1;
	ifc->pkt.fe[fn].cfg[0] = ifc->fcoe_cfg;
	fn++;

	//if (!priority_flow_control)
	//goto out;

	ifc->pkt.fe[fn].type = 3;
	ifc->pkt.fe[fn].oper_version = 0;
	ifc->pkt.fe[fn].max_version = 0;
	ifc->pkt.fe[fn].enable = 1;
	ifc->pkt.fe[fn].willing = 1;
	ifc->pkt.fe[fn].error = 0;
	ifc->pkt.fe[fn].sub_type = 0;
	ifc->pkt.fe[fn].cfg_size = 1;
	ifc->pkt.fe[fn].cfg[0] = ifc->priority_flow_control;
	fn++;

out:
	ifc->pkt.fe_size = fn;
}

/*******************************************************/
/*              Receive Packet                         */
/*******************************************************/

int get_type(unsigned char *buf)
{
	return buf[0]>>1;
}

int get_length(unsigned char *buf)
{
	return buf[1]+((buf[0]&0x1)<<8);
}

int for_each_tlv(struct interface_context *ifc, unsigned char *buf, int size, void(*parse_tlv)
                 (struct interface_context *ifc, int type, int length, unsigned char* value))
{
	int i;
	int length;
	int type;

	for (i = 0; i < size; i += length + 2) {
		type = get_type(buf + i);
		length = get_length (buf + i);
		parse_tlv(ifc, type, length, buf + i + 2);
	}
	return i;
}
		
struct feature_header {
	unsigned char oper_version;
	unsigned char max_version;
	int enable;
	int willing;
	int error;
	int sub_type;
};

void parse_feature_header(unsigned char *buf, struct feature_header *fh)
{
	fh->oper_version = buf[0];
	fh->max_version = buf[1];
	fh->enable = (buf[2] & 0x80)?1:0;
	fh->willing = (buf[2] & 0x40)?1:0;
	fh->error = (buf[2] & 0x20)?1:0;
	fh->sub_type = buf[3];
}

#define be32_to_cpu(x) (((x)[0]<<24) + ((x)[1]<<16) + ((x)[2]<<8) + ((x)[3]<<0))

void parse_protocol_control(unsigned char *buf, struct protocol_control *pc)
{
	pc->oper_version = buf[0];
	pc->max_version = buf[1];
	pc->seqno = be32_to_cpu (buf + 2);
	pc->ackno = be32_to_cpu (buf + 2 + 4);
}

void update_voltaile_fields_from_sub_tlv
(struct interface_context *ifc, int type, int length, unsigned char* value)
{
	struct protocol_control pc;
	struct feature_header fh;
	parse_feature_header(value, &fh);
	int new_state;

	if (type == 1) {
		parse_protocol_control(value, &pc);
		ifc->seqno = ((increment)?(pc.ackno + 1):1);
		ifc->ackno = pc.seqno;
	} else {				
		//general usage:
		//int cfg_length = length - 4
		//unsigned char *cfg = value + 4
		
		if ((type == 5) && (fh.sub_type == 0)) { //application fcoe
			if (ifc->fcoe_cfg != value[4]) {
				ifc->fcoe_cfg = value[4];
				ifc->dirty = 1;
				ifc->saw_an_fcoe_application_tlv = 1;
				if (verbose)
					printf("%s: changed fcoe priority map to %hhx\n", ifc->name, value[4]);
			}
		} else if (type == 3) {
			if (ifc->priority_flow_control != value[4]) {
				ifc->priority_flow_control = value[4];
				ifc->dirty = 1;
				if (verbose)
					printf("%s: changed priority flow control map to %hhx\n", ifc->name, value[4]);
			}			    
		} else if (monitor_link_state && (type == 6) && 
			   (fh.sub_type == 0)) { //fcoe logical link			
			new_state = (value[4] & 0x80)?1:0;
			change_link_state(ifc, new_state);
		}
	}
}

int is_dcbx(unsigned char *buf)
{
	return ((buf[0]==0x00)&&(buf[1]==0x1b)
		&&(buf[2]==0x21)&&(buf[3]==1));
}

void find_dcbx_tlv(struct interface_context *ifc, int type, int length, unsigned char* value)
{
	if ((type == 127) && (is_dcbx(value))) {
		for_each_tlv(ifc, value + 4, length - 4, 
			     update_voltaile_fields_from_sub_tlv);
	}
}

void parse_packet(unsigned char from[6], struct interface_context *ifc, unsigned char* buf, int size)
{
	for_each_tlv(ifc, buf, size, find_dcbx_tlv);
}

/*******************************************************/
/*                    print packet                     */
/*******************************************************/
char* subtype2name(int type) {
        switch (type) {
        case 1:
                return "Protocol_Control";
        case 2:
                return "Prioroty_Groups";
        case 3:
                return "Priority_Flow_Control";
        case 4:
                return "BCN";
        case 5:
                return "Application";
        case 6:
                return "Logical_Link_Down";
        }
        return "unknown_type";
}

void print_mac(unsigned char *buf)
{
        int i;
        for(i = 0; i < 6; i++) {
                printf("%.2x", buf[i]&0xff);
                if (i!=5) printf(":");
        }
}

void print_packet_dashbar()
{
        printf("-------------------------------\n");
}

void print_sub_tlv(struct interface_context *ifc, int type, int length, unsigned char *buf)
{
	struct feature_header fh;
	int show_cfg;

	printf("type: %d %s", type, subtype2name(type));
	
        if (type == 1) {
                struct protocol_control pc;
                parse_protocol_control(buf, &pc);
		printf("\n\tseqno: %u\n", pc.seqno);
		printf("\tackno: %u\n", pc.ackno);
        } else {
                parse_feature_header(buf, &fh);

                if((type == 5) && (fh.sub_type == 0))
                         printf(" [FCoE]\n");
                else if((type == 6) && (fh.sub_type == 0))
                         printf(" [FCoE]\n");
                else if((type == 6) && (fh.sub_type == 1))
                         printf(" [LAN]\n");
		else
			printf("\n");

		printf("\tenable: %d ", fh.enable);
		printf("willing: %d ", fh.willing);
		printf("error: %d\n", fh.error);
		printf("\tsub_type: %d\n", fh.sub_type);

                show_cfg = 1;
                switch (type) {
                case 5:
			printf("\tuser_priority_map: %d\n",buf[4]);
                        show_cfg = 0;
                        break;
                case 6:
			printf("\tlocal_link_status: %d\n",buf[4]>>7);
                        show_cfg = 0;
                        break;
                }

                if(show_cfg){
			printf("\tcfg: ");
                        print_hex(buf + 4, length - 4);
                }
        }
}

void print_from_lldp_tlv(struct interface_context *ifc, int type, int length, unsigned char *buf)
{	
	if (type != 127) return;
	if (!is_dcbx(buf)) return;
	for_each_tlv(ifc, buf + 4, length - 4, print_sub_tlv);
}

void print_packet(int is_incoming, unsigned char mac[6], struct interface_context *ifc,
                  unsigned char* buf, int size)
{
        print_packet_dashbar();
	printf("[%s] ", ifc->name);
        printf(is_incoming?"from: ":"to: ");
        print_mac(mac);
	printf("\n");
	for_each_tlv(ifc, buf, size, print_from_lldp_tlv);
        print_packet_dashbar();
}

/*******************************************************/
/*                      General                        */
/*******************************************************/
void update_packet(struct interface_context *ifc)
{
	ifc->pkt.pc.seqno = ifc->seqno;
	ifc->pkt.pc.ackno = ifc->ackno;
	ifc->pkt.fe[0].cfg[0] = ifc->fcoe_cfg;
	ifc->pkt.fe[1].cfg[0] = ifc->priority_flow_control;	
}

void update_vlan(struct interface_context *ifc)
{
	int prio;
	
	if (!ifc->announced_create && ifc->saw_an_fcoe_application_tlv) {
		for (prio = 0; prio < 8; prio++) {
			if (1 << prio & ifc->priority_flow_control)
				break;
		}
		
		ifc->prio = prio;
		create_vlan(ifc);
		if (!report_to_kernel)  
			ifc->announced_create = 1;
	}
}

void update_kernel(struct interface_context *ifc)
{
	char temp[50];
	int len;
	
	if (!ifc->announced_create && ifc->saw_an_fcoe_application_tlv) {
		len = sprintf(temp, "%s.0,%d", ifc->name, ifc->prio);
		write_buffer_to_sysfs(temp, len, "create");
		ifc->announced_create = 1;
	}
	if (ifc->tell_kernel_link_changed) {       
		len = sprintf(temp, "%s.0", ifc->name);
		write_buffer_to_sysfs(temp, len, ifc->link_state?"link_up":"link_down");
		ifc->tell_kernel_link_changed = 0;
	}
}

void handle_error(struct interface_context *ifc)
{
	change_link_state(ifc, 0);
	reinit_interface(ifc);
	ifc->link_state = 0;
}

int check_kernel_files()
{
	int ret;
	if ((ret = check_file("link_up")) < 0) return ret;
	if ((ret = check_file("link_down")) < 0) return ret;
	if ((ret = check_file("create")) < 0) return ret;
	return 0;
}

void usage(char *prog_name)
{
	printf("usage: %s [-d] [interfaces list]\n", prog_name);
	printf("e.g. %s eth2 eth4 eth6\n", prog_name);
	printf("\t-d disables reporting to the kernel via sysfs.\n");
	printf("\t-m monitor writing to sysfs.\n");
	printf("\t-q quiet mode.\n");
        printf("\t-v verbose: print progress.\n");
        printf("\t-vv very verbose: print packets.\n");
        printf("\t-s <sysfs base>. set sysfs path.\n");
	return;
}

void parse_args(int argc, char **argv)
{
        int i;
        opterr = 0;
	interface_count = 0;
	int c;

	int user_verbose = 0;
        int quiet = 0;

	path_to_sysfs = "/sys/class/mlx4_fc";

        while ((c = getopt (argc, argv, "s:dmqv")) != -1) {
                switch (c) {
                case 's':
			path_to_sysfs = optarg;
                        break;
                case 'd':
			report_to_kernel = 0;
                        break;
                case 'm':
			monitor_writing_to_sysfs = 1;
                        break;
                case 'q':
			quiet = 1;
                        break;
                case 'v':
                        user_verbose++;
                        break;
                case '?':
                        //if ((optopt == 'x'))
                        //        fprintf (stderr, "Option -%c requires an argument.\n", optopt);
                        //else if (isprint (optopt))
			if (isprint (optopt))
                                fprintf (stderr, "Unknown option `-%c'.\n", optopt);
                        else
                                fprintf (stderr, "Unknown option character `\\x%x'.\n", optopt);
                default:
                        usage(argv[0]);
			exit(1);
                }
	}
	 
        if (quiet)
                verbose = 0;
        else if (user_verbose)
                verbose = user_verbose;

        for (i = optind; i < argc; i++)
		interfaces[interface_count++].name = argv[i];

	if (interface_count == 0) {
		usage(argv[0]);
		exit(1);
	}
}

int main(int argv, char **argc)
{
	int i, rev;
	time_t now;
	parse_args(argv, argc);	
	if (init_interfaces() < 0) return 1;

	if (report_to_kernel) {
		init_sysfs_interface(path_to_sysfs);
		if (check_kernel_files() < 0) return 1;
	}

	for (i = 0; i < interface_count; i++) {
		memset(&(interfaces[i].pkt), 0, sizeof(struct packet));				
		init_packet(interfaces + i);
	}

	while (1) {
		recv_dcbx(parse_packet, handle_error);
		time(&now);	
		for(i = 0; i < interface_count; i++) {
			update_packet(interfaces + i);
			update_vlan(interfaces + i);
			if (report_to_kernel) update_kernel(interfaces + i);
			if ((interfaces[i].dirty) || (interfaces[i].expires < now)) {
				send_packet(interfaces + i);
				interfaces[i].dirty = 0;
				interfaces[i].expires = now + (interfaces[i].link_state?60:1);
			}
		}
	}	
	return 0;
}
