#!/bin/bash
#
# Copyright (c) 2006 Mellanox Technologies. All rights reserved.
#
# This Software is licensed under one of the following licenses:
#
# 1) under the terms of the "Common Public License 1.0" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/cpl.php.
#
# 2) under the terms of the "The BSD License" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/bsd-license.php.
#
# 3) under the terms of the "GNU General Public License (GPL) Version 2" a
#    copy of which is available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/gpl-license.php.
#
# Licensee has the right to choose one of the above licenses.
#
# Redistributions of source code must retain the above copyright
# notice and one of the license notices.
#
# Redistributions in binary form must reproduce both the above copyright
# notice, one of the license notices in the documentation
# and/or other materials provided with the distribution.


# Execute command w/ echo and exit if it fail
ex()
{
        echo "$@"
        if ! "$@"; then
                printf "\nFailed executing $@\n\n"
                exit 1
        fi
}

get_backports()
{
	echo 2.6.9_U4
	echo 2.6.9_U5
	echo 2.6.9_U6
	echo 2.6.9_U7
	echo 2.6.16_sles10
	echo 2.6.16_sles10_sp1
	echo 2.6.16_sles10_sp2
	echo 2.6.16
	echo 2.6.17
	echo 2.6.18_FC6 "(good for RHEL5 as well)"
	echo 2.6.18-EL5.1  "( RHEL5 U1)"
	echo 2.6.18-EL5.2  "( RHEL5 U2)"
	echo 2.6.18-EL5.3  "( RHEL5 U3)"
	echo 2.6.18_suse10_2
	echo 2.6.18
	echo 2.6.19
	echo 2.6.20
	echo 2.6.21
	echo 2.6.22
	echo 2.6.22_suse10_3
	echo 2.6.23
	echo 2.6.24
	echo 2.6.25
	echo 2.6.26
	echo 2.6.27
}

get_backport_dir()
{
    case $1 in
        2.6.9-42*)
                echo 2.6.9_U4
        ;;
        2.6.9-55*)
                echo 2.6.9_U5
        ;;
        2.6.9-67*)
                echo 2.6.9_U6
        ;;
	2.6.9-78*)
		echo 2.6.9_U7
	;;
        2.6.9*)
                echo
        ;;
        2.6.16.*-*-*)
                minor=$(echo $KVERSION | cut -d"." -f4 | cut -d"-" -f1)
                if [ $minor -lt 37 ]; then
                        echo 2.6.16_sles10
                elif [ $minor -lt 60 ]; then
                        echo 2.6.16_sles10_sp1
                else
                        echo 2.6.16_sles10_sp2
                fi
        ;;
        2.6.16*)
                echo 2.6.16
        ;;
        2.6.17*)
                echo 2.6.17
        ;;
        2.6.18-5[0-9]*.el5*)
                echo 2.6.18-EL5.1
        ;;
        2.6.18-9[0-9]*.el5*)
                echo 2.6.18-EL5.2
        ;;
        2.6.18-12[0-9]*.el5*)
                echo 2.6.18-EL5.3
        ;;
        2.6.18-*fc[56]*|2.6.18-*el5*)
                echo 2.6.18_FC6
        ;;
        2.6.18.*-*-*)
                middle=$(echo $KVERSION | cut -d"-" -f2)
                if [ $middle -lt 34 ]; then
                        echo 2.6.18
                else
                        echo 2.6.18_suse10_2
                fi
        ;;
        2.6.18*)
                echo 2.6.18
        ;;
        2.6.19*)
                echo 2.6.19
        ;;
        2.6.20*)
                echo 2.6.20
        ;;
        2.6.21*)
                echo 2.6.21
        ;;
        2.6.22.*-*-*)
                echo 2.6.22_suse10_3
        ;;
        2.6.22*)
                echo 2.6.22
        ;;
        2.6.23*)
                echo 2.6.23
        ;;
        2.6.24*)
                echo 2.6.24
        ;;
        2.6.25*)
                echo 2.6.25
        ;;
        2.6.26*)
                echo 2.6.26
        ;;
        2.6.27*)
                echo 2.6.27
        ;;
        *)
                echo
        ;;
    esac

}

# Apply patch
apply_patch()
{
        local patch=$1
        shift

        if [ -e  ${patch} ]; then
            printf "\t${patch}\n"
            if [ "${WITH_QUILT}" == "yes" ]; then
                ex $QUILT import ${patch}
                ex $QUILT push patches/${patch##*/}
            else
                if ! (patch -p1 -l < ${patch} ); then
                    echo "Failed to apply patch: ${patch}"
                    exit 1
                fi
            fi
        else
                echo File ${patch} does not exist
                return 1
        fi
        return 0
}


# Apply patches from the given directory
apply_backport_patches()
{
        local pdir=${CWD}/kernel_patches/backport/${BACKPORT_DIR}
        shift
        printf "\nApplying patches for ${BACKPORT_DIR} kernel:\n"
        if [ -d ${pdir} ]; then
                for patch in ${pdir}/*mlx4*
                do
                        apply_patch ${patch}
                done
        else
                echo ${pdir} no such directory
        fi
}

# Apply patches
patches_handle()
{
    ex mkdir -p ${CWD}/patches
    quiltrc=${CWD}/patches/quiltrc
    ex touch ${quiltrc}

cat << EOF >> ${quiltrc}
QUILT_DIFF_OPTS='-x .svn -p --ignore-matching-lines=\$Id'
QUILT_PATCH_OPTS='-l'
EOF

        QUILT="${QUILT} --quiltrc ${quiltrc}"

        # Apply backport patches
        BACKPORT_DIR=${BACKPORT_DIR:-$(get_backport_dir ${KVERSION})}
        if [ -n "${BACKPORT_DIR}" ]; then
               	apply_backport_patches
                BACKPORT_INCLUDES='-I${CWD}/kernel_addons/backport/'${BACKPORT_DIR}/include/
        fi
}

main() {

#Set default values
KVERSION=${KVERSION:-$(uname -r)}
WITH_QUILT=${WITH_QUILT:-"yes"}
WITH_PATCH=${WITH_PATCH:-"yes"}
BACKPORT_INCLUDES=""
EXTRA_FLAGS=""

WITH_BACKPORT_PATCHES=${WITH_BACKPORT_PATCHES:-"yes"}

QUILT=${QUILT:-$(/usr/bin/which quilt  2> /dev/null)}
CWD=$(pwd)
MEMTRACK_INCLUDE=""
if [ "$1" == "yes" ]; then
	MEMTRACK_INCLUDE="-include ${CWD}/drivers/net/debug/mtrack.h"
fi
CONFIG="config.mk"
PATCH_DIR=${PATCH_DIR:-""}

        # Check parameters
        if [ "$WITH_PATCH" == "yes" ] && [ "$WITH_QUILT" == "yes" ] && [[ ! -x ${QUILT} || ! -n "${QUILT}" ]]; then
                echo "Quilt ${QUILT} does not exist... Going to use patch."
                WITH_QUILT="no"
        fi

        patches_handle

# Remove patches dir
/bin/rm -rf ${CWD}/kernel_patches

        # Create config.mk
        /bin/rm -f ${CWD}/${CONFIG}
        cat >> ${CWD}/${CONFIG} << EOFCONFIG
KVERSION=${KVERSION}
ARCH=`uname -m`
MODULES_DIR:=/lib/modules/${KVERSION}/updates
KSRC:=/lib/modules/${KVERSION}/build
CWD=${CWD}
BACKPORT_INCLUDES:=${BACKPORT_INCLUDES}
MEMTRACK_INCLUDE:=${MEMTRACK_INCLUDE}
MLNX_EN_EXTRA_CFLAGS:=${EXTRA_FLAGS}
EOFCONFIG
        echo "Created ${CONFIG}:"
        cat ${CWD}/${CONFIG}

	# Create autoconf.h
	/bin/rm -f ${CWD}/include/linux/autoconf.h
	cat >> ${CWD}/include/linux/autoconf.h << EOFAUTO
#define CONFIG_MLX4_CORE 1
#define CONFIG_MLX4_EN 1
EOFAUTO
}

main $@
